'''

BINARY SEARCH TREE

Description:    Binary search tree implementation

'''


import copy
from abc import ABC, abstractmethod
from ..display import BinTreeDisplay

from .abstracttree import AbstractTree
from .binnode import BinNode
from .traversal import BinRecurTraversal
from .bintreeiter import BinTreeIterIn


#
#
class BinSearchTree(AbstractTree):
    '''
    Binary Search Tree.

    Args:
        lst: The list to construct from. If lst if None, the tree is empty.
        canddrm: Candidate chosen for replacement when doing "remove" action. This argument can be "left" or "right".
    '''
    #
    #
    #################################################################
    #                        METHODS (PUBLIC)
    #################################################################
    #
    #
    def __init__(self, lst: list = None, canddrm: str = 'right'):
        self._root = None

        self._size = 0

        if lst is not None and type(lst) is not list:
            raise ValueError('Invalid argument: lst must be a list')

        if canddrm not in ('left', 'right'):
            raise ValueError('Invalid argument: canddrm')

        self._option_candd_rm = canddrm

        if lst is not None:
            self._construct_from_list(lst)

    #
    #
    def empty(self):
        '''
        Checks if tree is empty.
        '''
        return self._root is None

    #
    #
    def size(self):
        '''
        Returns number of nodes.
        '''
        return self._size

    #
    #
    def height(self):
        '''
        Gets height of the tree.
        '''
        return self._height(self._root)

    #
    #
    def clear(self):
        '''
        Clears tree completely.
        '''
        self._dispose_root(self._root)
        self._root = None

    #
    #
    def traverse(self, order: str = 'in'):
        '''
        Traverses the tree entirely.
        Args:
            order: Type of traversal order.
                - 'pre': pre-order
                - 'in': in-order
                - 'post': post-order
        Returns:
            A list of keys.
        '''
        traversal = BinRecurTraversal()
        res = traversal.traverse(self._root, order)

        del traversal
        return res

    #
    #
    def contain(self, key):
        '''
        Checks if key exists.
        Args:
            key: The key to search for.
        Returns:
            True if key exists. Otherwise, None.
        '''
        if key is None:
            raise ValueError('Invalid argument: key must not be None')

        node, _ = self._search(self._root, key)
        return node is not None

    #
    #
    def insert(self, key):
        '''
        Inserts a key.
        Args:
            key: The key to insert.
        Returns:
            If insertion succeeds, return True. Otherwise, rerturn False.
        '''
        if key is None:
            raise ValueError('Invalid argument: key must not be None')

        self._success_state = False

        self._root = self._insert(self._root, key)

        if self._success_state is False:
            del self._success_state
            return False

        self._size += 1

        del self._success_state
        return True

    #
    #
    def remove(self, key):
        '''
        Removes the node of given key.
        Args:
            key: The key to be removed.
        Returns:
            If key exists, return True. Otherwise, rerturn False.
        '''
        if key is None:
            raise ValueError('Invalid argument: key must not be None')

        if self._root is None:
            return False

        if self.contain(key) is False:
            return False

        self._root = self._remove(self._root, key)

        self._size -= 1
        return True

    #
    #
    def min(self):
        '''
        Searches for minimum key.
        Args:
            None.
        Returns:
            Minimum key if exists. Otherwise, None.
        '''
        node, _ = self._search_min(self._root, None)
        res = node.key if node is not None else None
        return res

    #
    #
    def max(self):
        '''
        Searches for maximum key.
        Args:
            None.
        Returns:
            Maximum key if exists. Otherwise, None.
        '''
        node, _ = self._search_max(self._root, None)
        res = node.key if node is not None else None
        return res

    #
    #
    #################################################################
    #                        METHODS (PROTECTED) - BASE BINARY TREE
    #################################################################
    #
    #
    def _height(self, node: BinNode):
        if node is None:
            return 0

        height_le = self._height(node.left)
        height_ri = self._height(node.right)

        return 1 + max(height_le, height_ri)

    #
    #
    def _create_node(self, key=None):
        return BinNode(key)

    #
    #
    def _dispose_root(self, node: BinNode):
        '''
        Helps garbage collection free memory used by tree.
        Args:
            node: Input node.
        '''
        if node is None:
            return

        self._dispose_root(node.left)
        del node.left

        self._dispose_root(node.right)
        del node.right

        del node.key

    #
    #
    #################################################################
    #                        METHODS (PROTECTED)
    #################################################################
    #
    #
    def _insert(self, node: BinNode, key):
        '''
        Inserts a key (backend function)
        Args:
            node: Current processing node.
            key: The key to insert.
        Returns:
            The current processing node itself.
        '''
        if node is None:
            self._success_state = True
            return self._create_node(key)

        if key < node.key:
            node.left = self._insert(node.left, key)
        elif key > node.key:
            node.right = self._insert(node.right, key)

        return node

    #
    #
    def _remove(self, node: BinNode, key):
        '''
        Removes the node from the tree.
        Args:
            node: The current processing node.
            key: The key to search and remove.
        Returns:
            The current processing node itself.
        '''
        if node is None:
            return None

        if key < node.key:
            node.left = self._remove(node.left, key)
        elif key > node.key:
            node.right = self._remove(node.right, key)
        else:
            if node.left is None:
                return node.right

            if node.right is None:
                return node.left

            self._remove_candidate(node)

        return node

    #
    #
    def _remove_candidate(self, node: BinNode):
        '''
        Finds a candidate for replacement of current node, and then removes that candidate from the tree.
        Args:
            node: The starting node.
        '''
        assert node is not None

        option = self._option_candd_rm

        if option == 'right':
            candidate, _ = self._search_min(node.right, node)
            node.key = candidate.key
            node.right = self._remove(node.right, candidate.key)

        elif option == 'left':
            candidate, _ = self._search_max(node.left, node)
            node.key = candidate.key
            node.left = self._remove(node.left, candidate.key)

    #
    #
    def _search(self, node: BinNode, key):
        '''
        Searches for a key.
        Args:
            key: The key to search for.
        Returns:
            The tuple (node, parent) indicating result node and its parent.
        '''
        parent = None

        while True:
            if node is None:
                return (None, None)

            if key == node.key:
                return (node, parent)

            parent = node

            if key < node.key:
                node = node.left
            else:
                node = node.right

        return (None, None)  # ensure a value to return, unreachable statement...

    #
    #
    def _search_min(self, node: BinNode, parent: BinNode):
        '''
        Searches for the node with minimum key.
        Args:
            node: Starting node.
            parent: Parent of starting node.
        Returns:
            The tuple (node, parent) indicating result node and its parent.
        '''
        if node is None:
            return (None, parent)

        while node.left is not None:
            parent = node
            node = node.left

        return (node, parent)

    #
    #
    def _search_max(self, node: BinNode, parent: BinNode):
        '''
        Searches for the node with maximum key.
        Args:
            node: Starting node.
            parent: Parent of starting node.
        Returns:
            The tuple (node, parent) indicating result node and its parent.
        '''
        if node is None:
            return (None, parent)

        while node.right is not None:
            parent = node
            node = node.right

        return (node, parent)

    #
    #
    def _construct_from_list(self, lst: list):
        '''
        Constructs tree from a list. WARNING: The list will change its value (i.e. sorting).
        Args:
            lst: The list.
        Returns:
            None.
        '''
        self._dispose_root(self._root)
        self._root = None

        lst = sorted(set(lst))
        len_lst = len(lst)

        self._root = self._build_tree_from_sorted_list(lst, 0, len_lst - 1)
        self._size = len_lst

    #
    #
    def _build_tree_from_sorted_list(self, lst: list, index_start: int, index_end: int):
        '''
        Backend function for "__construct_from_list".
        Args:
            lst: The list.
            index_start: Starting index in list.
            index_end: Ending index in list (inclusive).
        Returns:
            Root node (followed by its branch) constructed from lst[ index_start to index_end ]
        '''
        if index_start > index_end:
            return None

        index_mid = (index_start + index_end) // 2
        node = self._create_node(lst[index_mid])

        node.left = self._build_tree_from_sorted_list(lst, index_start, index_mid - 1)
        node.right = self._build_tree_from_sorted_list(lst, index_mid + 1, index_end)

        self._build_tree_from_sorted_list_node_func(node)

        return node

    #
    #
    def _build_tree_from_sorted_list_node_func(self, node):
        pass

    #
    #
    #################################################################
    #                        METHODS (EXTRA)
    #################################################################
    #
    #
    # Adapter method connecting BinTreeDisplay
    def display(self, disp: BinTreeDisplay) -> str:
        res = disp.get(self._root)
        return res

    #
    #
    # Adapter method connecting BinTreeDisplay
    def display_lst_rows(self, disp: BinTreeDisplay) -> list:
        res = disp.get_lst_rows(self._root)
        return res

    #
    #
    def __str__(self):
        return 'BST' + self._get_traversal_str()



from .itraversal import ITraversal


#
#
class BinRecurTraversal(ITraversal):
    #
    #
    def traverse(self, root, order: str) -> list:
        if order not in ('pre', 'in', 'post'):
            raise ValueError('Invalid argument: order')

        self._res_path = []

        if order == 'pre':
            self._traverse_pre(root)

        elif order == 'in':
            self._traverse_in(root)

        elif order == 'post':
            self._traverse_post(root)

        res = self._res_path
        del self._res_path

        return res

    #
    #
    def _traverse_pre(self, node):
        if node is None:
            return

        self._res_path.append(node.key)
        self._traverse_pre(node.left)
        self._traverse_pre(node.right)

    #
    #
    def _traverse_in(self, node):
        if node is None:
            return

        self._traverse_in(node.left)
        self._res_path.append(node.key)
        self._traverse_in(node.right)

    #
    #
    def _traverse_post(self, node):
        if node is None:
            return

        self._traverse_post(node.left)
        self._traverse_post(node.right)
        self._res_path.append(node.key)



class BinNode:
    '''
    Binary Node.

    Args:
        key: The key stored in the node.
    '''
    #
    #
    def __init__(self, key=None):
        self.key = key
        self.left = None
        self.right = None

    #
    #
    def __str__(self):
        return str(self.key)



class ValueUtil:
    #
    #
    def __init__(self, float_precision: int = 2):
        self.set_float_precision(float_precision)

    #
    #
    def get_float_precision(self):
        return self.__float_precision

    #
    #
    def set_float_precision(self, precision: int):
        if type(precision) is not int or precision < 0:
            raise ValueError('Invalid argument: float precision must be a non-negative integer')

        self.__float_precision = precision

    #
    #
    def get_str(self, value) -> str:
        type_value = type(value)

        if type_value is float:
            return str(round(value, self.__float_precision))

        if type_value is str:
            return value

        return str(value)



class MatrixBuffer:
    '''
    Matrix buffer which stores an ASCII image.
    '''
    #
    #
    def __init__(self, width: int, height: int):
        if type(width) is not int or width < 1:
            raise ValueError('width must be a positive integer')

        if type(height) is not int or height < 1:
            raise ValueError('height must be a positive integer')

        self.__width = width
        self.__height = height

        self.__a = [[' '] * width for _ in range(height)]

        self.__y_bottom = -1

    #
    #
    def width(self):
        return self.__width

    #
    #
    def height(self):
        return self.__height

    #
    #
    def fill_str(self, x: int, y: int, value: str):
        if type(value) is not str:
            raise ValueError('Invalid argument: value must be a string')

        self.__check_arg_x(x, 'x')
        self.__check_arg_y(y, 'y')

        a = self.__a
        x_start = x
        len_value = len(value)

        for i in range(len_value):
            x = x_start + i

            if x >= self.__width:
                break

            a[y][x] = value[i]

        self.__update_y_bottom(y)

    #
    #
    def fill_hori_line(self, character: str, y: int, x_start: int, x_end: int):
        self.__check_arg_y(y, 'y')
        self.__check_arg_x(x_start, 'x_start')
        self.__check_arg_x(x_end, 'x_end')

        for x in range(x_start, x_end + 1):
            self.__a[y][x] = character

        self.__update_y_bottom(y)

    #
    #
    def fill_vert_line(self, character: str, x: int, y_start: int, y_end: int):
        self.__check_arg_x(x, 'x')
        self.__check_arg_y(y_start, 'y_start')
        self.__check_arg_y(y_end, 'y_end')

        for y in range(y_start, y_end + 1):
            self.__a[y][x] = character

        self.__update_y_bottom(y_end)

    #
    #
    def __check_arg_x(self, value: int, name_arg: str):
        if value < 0 or value >= self.__width:
            raise ValueError('Invalid argument:', name_arg)

    #
    #
    def __check_arg_y(self, value: int, name_arg: str):
        if value < 0 or value >= self.__height:
            raise ValueError('Invalid argument:', name_arg)

    #
    #
    def __update_y_bottom(self, value: int):
        self.__y_bottom = max(self.__y_bottom, value)

    #
    #
    def get_str(self) -> str:
        lst_rows = self.get_lst_rows()
        res = '\n'.join(lst_rows)
        return res

    #
    #
    def get_lst_rows(self) -> list:
        lst_rows = [''.join(row) for row in self.__a]

        # optimization: remove unnecessary ending row
        del lst_rows[self.__y_bottom + 1:]

        return lst_rows



from collections import deque
from .itraversal import ITraversal


#
#
class BinNonRecurTraversal(ITraversal):
    #
    #
    def traverse(self, root, order: str) -> list:
        if order not in ('pre', 'in', 'post'):
            raise ValueError('Invalid argument: order')

        res_path = []

        if root is None:
            return res_path

        if order == 'pre':
            res_path = self._traverse_pre(root)

        elif order == 'in':
            res_path = self._traverse_in(root)

        elif order == 'post':
            res_path = self._traverse_post(root)

        return res_path

    #
    #
    def _traverse_pre(self, node):
        res_path = []
        stack = deque()

        stack.append(node)

        while stack:
            node = stack.pop()
            res_path.append(node.key)

            if node.right is not None:
                stack.append(node.right)

            if node.left is not None:
                stack.append(node.left)

        return res_path

    #
    #
    def _traverse_post(self, node):
        res_path = []
        stack = deque()

        stack.append(node)

        while stack:
            node = stack.pop()
            res_path.append(node.key)

            if node.left is not None:
                stack.append(node.left)

            if node.right is not None:
                stack.append(node.right)

        return res_path[::-1]

    #
    #
    def _traverse_in(self, node):
        res_path = []
        stack = deque()

        while True:
            while node is not None:
                stack.append(node)
                node = node.left

            if not stack:
                break

            picked_node = stack.pop()
            res_path.append(picked_node.key)

            node = picked_node.right

        return res_path



from abc import ABC, abstractmethod
from .valueutil import ValueUtil
from .matrixbuffer import MatrixBuffer
from .parsingnode import ParsingNode


#
#
class AbstractTreeDisplay(ABC):
    '''
    Base class for tree display.
    '''
    #
    #
    #################################################################
    #                        METHODS (PUBLIC)
    #################################################################
    #
    #
    def __init__(self):
        self._vutil = ValueUtil()
        self._parser = None
        self._buffer = None

        self._config(dict(
            leaf_at_bottom=False, compact_vert_line=False, margin_left=0, hori_line_char='-'
        ))

    #
    #
    def get(self, inp_root) -> str:
        '''
        Gets ASCII display string of tree. Output result is configurable by calling method "config".

        Args:
            inp_root: Input root of tree.

        Returns:
            String result. If inp_root is None then returns an empty string.
        '''
        if inp_root is None:
            return ''

        self._process(inp_root)
        assert self._buffer is not None

        res = self._buffer.get_str()

        del self._buffer
        return res

    #
    #
    def get_lst_rows(self, inp_root) -> list:
        '''
        Gets ASCII display string of tree. Output result is configurable by calling method "config".

        Args:
            inp_root: Input root of tree.

        Returns:
            List of rows. Each row is a string. If inp_root is None then returns an empty list.
        '''
        if inp_root is None:
            return []

        self._process(inp_root)
        assert self._buffer is not None

        res = self._buffer.get_lst_rows()

        del self._buffer
        return res

    #
    #
    #################################################################
    #                        METHODS (PROTECTED)
    #################################################################
    #
    #
    def _config(self, kwargs):
        '''
        Configures settings:
            struct_node, space_branch_neighbor, float_pre, leaf_at_bottom,
            compact_vert_line, margin_left, hori_line_char
        '''
        #
        argval = kwargs.pop('struct_node', None)
        if argval is not None:
            self._parser.config(struct_node=argval)

        #
        argval = kwargs.pop('space_branch_neighbor', None)
        if argval is not None:
            self._parser.config(space_branch_neighbor=argval)

        #
        argval = kwargs.pop('float_pre', None)
        if argval is not None:
            self._vutil.set_float_precision(argval)

        #
        argval = kwargs.pop('leaf_at_bottom', None)
        if argval is not None:
            if type(argval) is not bool:
                raise ValueError('Invalid argument: leaf_at_bottom must be a boolean')

            self._leaf_at_bottom = argval

        #
        argval = kwargs.pop('compact_vert_line', None)
        if argval is not None:
            if type(argval) is not bool:
                raise ValueError('Invalid argument: compact_vert_line must be boolean')

            self._compact_vert_line = argval

        #
        argval = kwargs.pop('margin_left', None)
        if argval is not None:
            if type(argval) is not int or argval < 0:
                raise ValueError('Invalid argument: margin_left must be a non-negative integer')

            self._margin_left = argval

        #
        argval = kwargs.pop('hori_line_char', None)
        if argval is not None:
            if type(argval) is not str or len(argval) != 1:
                raise ValueError('Invalid argument: hori_line_char must be a string of length 1')

            self._hori_line_char = argval

    #
    #
    def _process(self, inp_root):
        '''
        Backend function for "get" method.
        '''
        if inp_root is None:
            raise ValueError('Cannot process an empty tree')

        height_inp_root = self._parser.get_height(inp_root)
        height_buffer = self._get_height_buffer(height_inp_root)

        parsing_tree = self._parser.build_tree(inp_root)
        self._parser.convert_margin_local_to_global(parsing_tree, self._margin_left)

        self._buffer = MatrixBuffer(parsing_tree.width + self._margin_left, height_buffer)
        self._fill_buffer(parsing_tree, 0)

        self._parser.destroy_tree(parsing_tree)

    #
    #
    def _fill_buffer_node_leaf(self, node: ParsingNode, y_depth: int):
        y_depth_end = y_depth

        if self._leaf_at_bottom:
            y_depth_end = self._buffer.height() - 1
            self._buffer.fill_vert_line('|', node.margin_key_center, y_depth, y_depth_end - 1)

        self._buffer.fill_str(node.margin_key, y_depth_end, node.key)

    #
    #
    @abstractmethod
    def _fill_buffer(self, node: ParsingNode, y_depth: int):
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def _get_height_buffer(self, height_inp_root: int) -> int:
        raise NotImplementedError()



from .abtreedisplay import AbstractTreeDisplay
from .valueutil import ValueUtil
from .matrixbuffer import MatrixBuffer
from .parsingnode import ParsingNode
from .nonbintreeparser import NonBinTreeParser


#
#
class NonBinTreeDisplay(AbstractTreeDisplay):
    '''
    A utility help visualize non-binary trees by using ASCII text.
    '''
    #
    #
    #################################################################
    #                        METHODS (PUBLIC)
    #################################################################
    #
    #
    def __init__(self):
        super().__init__()
        self._parser = NonBinTreeParser(self._vutil)
        self.config(compact_single_leaf=True)

    #
    #
    def config(self, **kwargs):
        '''
        Configures settings.

        Args:
            struct_node: Tuple(name_key, name_lst_children) indicating structure information of input node.

            space_branch_neighbor: Space between 2 branch neighbors.

            float_pre: Maximum precision of floating-point numbers.

            leaf_at_bottom: True if leaf will be drawn at bottom. Otherwise, False.

            compact_vert_line: True if vertical dash below will be omitted. Otherwise, False.

            margin_left: Left margin of output string result.

            hori_line_char: Display character for the horizontal line connecting branches.

            compact_single_leaf: True, if node has only a single child then this child will be compact. Otherwise, False. This setting is disabled when leaf_at_bottom is True.
        '''
        super()._config(kwargs)

        #
        argval = kwargs.pop('compact_single_leaf', None)
        if argval is not None:
            if type(argval) is not bool:
                raise ValueError('Invalid argument: compact_single_leaf must be a boolean')

            self._compact_single_leaf = argval

        #
        if kwargs:
            args_invalid = list(kwargs.keys())
            raise ValueError('Invalid argument(s):', args_invalid)

    #
    #
    #################################################################
    #                        METHODS (PROTECTED)
    #################################################################
    #
    #
    def _fill_buffer(self, node: ParsingNode, y_depth: int):
        if node is None:
            return

        if node.is_leaf:
            self._fill_buffer_node_leaf(node, y_depth)
            return

        num_children = len(node.children)

        self._buffer.fill_str(node.margin_key, y_depth, node.key)
        self._buffer.fill_str(node.margin_key_center, y_depth + 1, '|')

        if num_children == 1 and not self._leaf_at_bottom and self._compact_single_leaf:
            self._buffer.fill_str(node.children[0].margin_key, y_depth + 2, node.children[0].key)
            return

        hori_line_char = self._hori_line_char if num_children > 1 else '|'
        self._buffer.fill_hori_line(hori_line_char, y_depth + 2, node.hori_line_xstart, node.hori_line_xend)

        y_depth += 3
        y_depth += 1 if not self._compact_vert_line else 0

        for i in range(num_children):
            if not self._compact_vert_line:
                self._buffer.fill_str(node.margin_vert_dash_below[i], y_depth - 1, '|')
            self._fill_buffer(node.children[i], y_depth)

    #
    #
    def _get_height_buffer(self, height_inp_root: int) -> int:
        if self._compact_vert_line:
            return height_inp_root * 3 - 2

        return height_inp_root * 4 - 3



from .parsingnode import ParsingNode
from .valueutil import ValueUtil


#
#
class BinTreeParser:
    #
    #
    def __init__(self, value_util: ValueUtil):
        if value_util is None:
            raise ValueError('value_util cannot be None')

        self.vutil = value_util
        self.config(struct_node=('key', 'left', 'right'), space_branch_neighbor=3)

    #
    #
    def config(self, **kwargs):
        '''
        Configures settings.

        Args:
            struct_node: Tuple(name_key, name_left, name_right) indicating structure information of input node.

            space_branch_neighbor: Space between 2 branch neighbors.
        '''
        for arg, argval in kwargs.items():
            if arg == 'struct_node':
                if type(argval) is not tuple or len(argval) != 3 \
                        or type(argval[0]) is not str or type(argval[1]) is not str or type(argval[2]) is not str:
                    raise ValueError('Invalid argument: struct_node must be a tuple(name_key, name_left, name_right)')

                self.struct_node_key = argval[0]
                self.struct_node_le = argval[1]
                self.struct_node_ri = argval[2]

            elif arg == 'space_branch_neighbor':
                if type(argval) is not int or argval < 1:
                    raise ValueError('Invalid argument: space_branch_neighbor must be a positive integer')

                self.space_branch_neighbor = argval

            else:
                raise ValueError('Invalid argument:', arg)

    #
    #
    def build_tree(self, input_root) -> ParsingNode:
        '''
        Builds parsing tree which stores parsing information of each corresponding node.
        The structure of input_root should be configured by method "config".

        Args:
            input_root: Input root node.

        Returns:
            A node of type ParsingNode indicating parsing tree.
        '''
        if input_root is None:
            return None

        input_key = getattr(input_root, self.struct_node_key)
        input_le = getattr(input_root, self.struct_node_le)
        input_ri = getattr(input_root, self.struct_node_ri)

        node = ParsingNode(self.vutil.get_str(input_key))

        # STEP 1. Children tree processing
        node.le = self.build_tree(input_le)
        node.ri = self.build_tree(input_ri)

        # STEP 2. Common factors
        has_child_le = node.le is not None
        has_child_ri = node.ri is not None
        is_leaf = not has_child_le and not has_child_ri

        width_child_le = node.le.width if has_child_le else 0
        width_child_ri = node.ri.width if has_child_ri else 0

        margin_child_ri = width_child_le + self.space_branch_neighbor

        margin_vert_dash_below_le = node.le.margin_key_center if has_child_le else 0
        margin_vert_dash_below_ri = margin_child_ri + (node.ri.margin_key_center if has_child_ri else 0)

        # STEP 3. Horizontal line
        hori_line_xstart, hori_line_xend = None, None

        if not is_leaf:
            hori_line_xstart = margin_vert_dash_below_le
            hori_line_xend = margin_vert_dash_below_ri

            if has_child_le and not has_child_ri:
                hori_line_xend -= 1

        # STEP 4. Margin of current node
        margin_key_center = tmp = max(0, len(node.key) - 1) // 2

        if not is_leaf:
            if has_child_le and has_child_ri:
                margin_key_center = (hori_line_xstart + hori_line_xend) // 2
            elif has_child_le:
                margin_key_center = hori_line_xend
            elif has_child_ri:
                margin_key_center = 0

        margin_key = margin_key_center - tmp

        # STEP 5. Width of current node
        width_chbrsp = margin_child_ri + width_child_ri if not is_leaf else 0
        full_width = width_chbrsp if not is_leaf else len(node.key)

        # FINAL STEP
        node.width = full_width
        node.margin_key = margin_key
        node.margin_key_center = margin_key_center
        node.is_leaf = is_leaf

        if not is_leaf:
            node.width_chbrsp = width_chbrsp  # width of children + branch spacing
            node.hori_line_xstart = hori_line_xstart
            node.hori_line_xend = hori_line_xend
            node.margin_vert_dash_below_le = margin_vert_dash_below_le
            node.margin_vert_dash_below_ri = margin_vert_dash_below_ri
            node.margin_child_le = 0
            node.margin_child_ri = margin_child_ri

        self.adjust_margin_key(node)

        del input_key
        del input_le
        del input_ri
        del has_child_le
        del has_child_ri
        del is_leaf
        del width_child_le
        del width_child_ri
        del margin_child_ri
        del margin_vert_dash_below_le
        del margin_vert_dash_below_ri
        del hori_line_xstart
        del hori_line_xend
        del margin_key_center
        del margin_key
        del width_chbrsp
        del full_width
        del tmp

        return node

    #
    #
    def adjust_margin_key(self, node: ParsingNode):
        len_key = len(node.key)

        if node.margin_key >= 0 and node.margin_key + len_key <= node.width:
            return

        shift_factor = 0 if node.margin_key >= 0 else abs(node.margin_key)

        node.width = max(node.width, node.margin_key + len_key)
        node.width += shift_factor

        self.shift_margin_except_children(node, shift_factor)

        del shift_factor

    #
    #
    def convert_margin_local_to_global(self, node: ParsingNode, shift_factor: int):
        self.shift_margin_except_children(node, shift_factor)

        if node.le is not None:
            self.convert_margin_local_to_global(node.le, node.margin_child_le)

        if node.ri is not None:
            self.convert_margin_local_to_global(node.ri, node.margin_child_ri)

    #
    #
    def shift_margin_except_children(self, node: ParsingNode, shift_factor: int):
        node.margin_key += shift_factor
        node.margin_key_center += shift_factor

        if not node.is_leaf:
            node.hori_line_xstart += shift_factor
            node.hori_line_xend += shift_factor
            node.margin_vert_dash_below_le += shift_factor
            node.margin_vert_dash_below_ri += shift_factor
            node.margin_child_le += shift_factor
            node.margin_child_ri += shift_factor

    #
    #
    def destroy_tree(self, node: ParsingNode):
        if node is None:
            return

        self.destroy_tree(node.le)
        del node.le

        self.destroy_tree(node.ri)
        del node.ri

        del node.key
        del node.width
        del node.margin_key
        del node.margin_key_center

        if not node.is_leaf:
            del node.width_chbrsp
            del node.hori_line_xstart
            del node.hori_line_xend
            del node.margin_vert_dash_below_le
            del node.margin_vert_dash_below_ri
            del node.margin_child_le
            del node.margin_child_ri

        del node.is_leaf

    #
    #
    def get_height(self, node) -> int:
        if node is None:
            return 0

        height_le = self.get_height(getattr(node, self.struct_node_le))
        height_ri = self.get_height(getattr(node, self.struct_node_ri))

        return 1 + max(height_le, height_ri)



from .binnode import BinNode


#
#
class AvlNode(BinNode):
    '''
    Avl Node.

    Args:
        key: The key stored in the node.
    '''
    #
    #
    def __init__(self, key=None):
        super().__init__(key)
        self._height = 1

    #
    #
    def height(self):
        return self._height

    #
    #
    def height_left(self):
        return 0 if self.left is None else self.left.height()

    #
    #
    def height_right(self):
        return 0 if self.right is None else self.right.height()

    #
    #
    def update_height(self):
        self._height = 1 + max(self.height_left(), self.height_right())

    #
    #
    def balance(self):
        '''
        Gets balance factor of the node.
        '''
        height_le = self.height_left()
        height_ri = self.height_right()
        return height_le - height_ri

    #
    #
    def balance_left(self):
        '''
        Gets balance factor of the left branch.
        '''
        return 0 if self.left is None else self.left.balance()

    #
    #
    def balance_right(self):
        '''
        Gets balance factor of the right branch.
        '''
        return 0 if self.right is None else self.right.balance()



'''

ABSTRACT TREE

Description:    Abstract base class of a tree

'''


import copy
from abc import ABC, abstractmethod


#
#
class AbstractTree(ABC):
    '''
    Abstract base class of a tree.
    '''
    #################################################################
    #                        ABSTRACT METHODS
    #################################################################
    #
    #
    def __init__(self):
        self._root = None

    #
    #
    @abstractmethod
    def empty(self):
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def size(self):
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def height(self):
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def clear(self):
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def traverse(self, order: str = 'in') -> list:
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def insert(self, key):
        raise NotImplementedError()

    #
    #
    @abstractmethod
    def remove(self, key):
        raise NotImplementedError()

    #
    #
    #################################################################
    #                        METHODS (PUBLIC)
    #################################################################
    #
    #
    def clone(self):
        '''
        Clones tree completely.
        '''
        the_clone = copy.deepcopy(self)
        return the_clone

    #
    #
    #################################################################
    #                        METHODS (EXTRA)
    #################################################################
    #
    #
    def _get_traversal_str(self) -> str:
        lst_traversal = self.traverse('in')
        res = '  '.join(str(x) for x in lst_traversal)
        res = f'({res})'
        return res

    #
    #
    def __str__(self) -> str:
        return self._get_traversal_str()



'''

AVL TREE

Description:    AVL tree implementation

'''


from .binsearchtree import BinSearchTree
from .avlnode import AvlNode


#
#
class AvlTree(BinSearchTree):
    '''
    AVL Tree.

    Args:
        lst: The list to construct from. If lst if None, the tree is empty.
        canddrm: Candidate chosen for replacement when doing "remove" action. This argument can be "left" or "right".
    '''
    #
    #
    #################################################################
    #                        METHODS (PUBLIC)
    #################################################################
    #
    #
    def __init__(self, lst: list = None, canddrm: str = 'right'):
        super().__init__(lst, canddrm)

    #
    #
    def height(self):
        '''
        Gets height of the tree.
        '''
        if self._root is None:
            return 0

        return self._root.height()

    #
    #
    #################################################################
    #                        METHODS (PROTECTED)
    #################################################################
    #
    #
    def _insert(self, node: AvlNode, key):
        '''
        Inserts a key (backend function)
        Args:
            node: Current processing node.
            key: The key to insert.
        Returns:
            The current processing node itself.
        '''
        node = super()._insert(node, key)
        node = self._adjust_balance(node)
        return node

    #
    #
    def _remove(self, node: AvlNode, key):
        '''
        Removes the node from the tree.
        Args:
            node: The current processing node.
            key: The key to search and remove.
        Returns:
            The current processing node itself.
        '''
        node = super()._remove(node, key)

        if node is not None:
            node = self._adjust_balance(node)

        return node

    #
    #
    def _adjust_balance(self, node: AvlNode):
        # STEP 1. Update the height of the node
        node.update_height()

        # STEP 2. Get the balance factor
        balance = node.balance()
        balance_le = node.balance_left()
        balance_ri = node.balance_right()

        # STEP 3. Process if the node is unbalanced ==> 4 cases

        # Case 1: left-left
        if balance > 1 and balance_le >= 0:
            return self._rotate_right(node)

        # Case 2: right-right
        if balance < -1 and balance_ri <= 0:
            return self._rotate_left(node)

        # Case 3: left-right
        if balance > 1 and balance_le < 0:
            node.left = self._rotate_left(node.left)
            return self._rotate_right(node)

        # Case 4: right-left
        if balance < -1 and balance_ri > 0:
            node.right = self._rotate_right(node.right)
            return self._rotate_left(node)

        return node

    #
    #
    def _rotate_left(self, node: AvlNode):
        r'''
            node
            /  \
           ..   V
               / \
              a   b
        '''
        V = node.right
        a = V.left

        V.left = node
        node.right = a

        node.update_height()
        V.update_height()

        return V

    #
    #
    def _rotate_right(self, node: AvlNode):
        r'''
            node
            /  \
           V   ..
          / \
         a   b
        '''
        V = node.left
        b = V.right

        V.right = node
        node.left = b

        node.update_height()
        V.update_height()

        return V

    #
    #
    def _create_node(self, key=None):
        return AvlNode(key)

    #
    #
    def _build_tree_from_sorted_list_node_func(self, node: AvlNode):
        node.update_height()

    #
    #
    ################################################################
    #                        METHOD (EXTRA)
    ################################################################
    #
    #
    def __str__(self):
        return 'AVL' + self._get_traversal_str()



'''

TREE TRAVERSAL

Description:    Base class which help traverse through a tree.

'''


#
#
class ITraversal:
    def traverse(self, root, order: str) -> list:
        '''
        Traverses a tree entirely.
        Args:
            order: Type of traversal order.
                - 'pre': pre-order
                - 'in': in-order
                - 'post': post-order
        Returns:
            A list of keys.
        '''
        raise NotImplementedError()



'''
TREE DISPLAY LIBRARY

Description:    A utility help visualize trees by using ASCII text.
'''


from .nonbintreedisplay import NonBinTreeDisplay
from .bintreedisplay import BinTreeDisplay


__all__ = [
    'NonBinTreeDisplay',
    'BinTreeDisplay'
]



from setuptools import setup, find_packages

with open('README.md') as file_readme:
    README = file_readme.read()

setup_args = dict(
    name='sutree',
    version='0.2.0',
    description='Tree data structure library',
    long_description_content_type="text/markdown",
    long_description=README + '\n',
    license='3-Clause BSD License',
    packages=find_packages(exclude=['*test.*', '*test']),
    author='Thanh Trung Nguyen',
    author_email='thanh.it1995@gmail.com',

    keywords=['tree', 'binary', 'avl', 'bst',
              'display', 'view', 'print', 'visualize', 'visualization', 'ascii', 'console', 'terminal'],

    url='https://github.com/thanhit95/sutree',
    download_url='https://pypi.org/project/sutree'
)

install_requires = [
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)



from .parsingnode import ParsingNode
from .valueutil import ValueUtil


#
#
class NonBinTreeParser:
    #
    #
    def __init__(self, value_util: ValueUtil):
        if value_util is None:
            raise ValueError('value_util cannot be None')

        self.vutil = value_util
        self.config(struct_node=('key', 'children'), space_branch_neighbor=3)

    #
    #
    def config(self, **kwargs):
        '''
        Configures settings.

        Args:
            struct_node: Tuple(name_key, name_lst_children) indicating structure information of input node.

            space_branch_neighbor: Space between 2 branch neighbors.
        '''
        for arg, argval in kwargs.items():
            if arg == 'struct_node':
                if type(argval) is not tuple or len(argval) != 2 \
                        or type(argval[0]) is not str or type(argval[1]) is not str:
                    raise ValueError('Invalid argument: struct_node must be a tuple(name_key, name_children)')

                self.struct_node_key = argval[0]
                self.struct_node_children = argval[1]

            elif arg == 'space_branch_neighbor':
                if type(argval) is not int or argval < 1:
                    raise ValueError('Invalid argument: space_branch_neighbor must be a positive integer')

                self.space_branch_neighbor = argval

            else:
                raise ValueError('Invalid argument:', arg)

    #
    #
    def build_tree(self, input_root) -> ParsingNode:
        '''
        Builds parsing tree which stores parsing information of each corresponding node.
        The structure of input_root should be configured by method "config".

        Args:
            input_root: Input root node.

        Returns:
            A node of type ParsingNode indicating parsing tree.
        '''
        if input_root is None:
            return None

        input_key = getattr(input_root, self.struct_node_key)
        input_children = getattr(input_root, self.struct_node_children)

        node = ParsingNode(self.vutil.get_str(input_key))
        node.children = []

        # STEP 1. Children tree processing
        for input_child in input_children:
            if input_child is None:
                continue
            parsing_child_node = self.build_tree(input_child)
            node.children.append(parsing_child_node)

        children = node.children  # alias

        # STEP 2. Common factors
        num_children = len(children)
        is_leaf = (num_children == 0)

        width_children = [child.width for child in children]

        # magic adjustment ==> the most left child
        if num_children >= 2:
            tmp = len(children[0].key)
            if tmp > 0 and tmp % 2 == 0:
                children[0].margin_key_center += 1

        margin_prefix_children = self.get_margin_prefix_children(num_children, width_children)
        margin_vert_dash_below = self.get_margin_vert_dash_below(children, margin_prefix_children)

        # STEP 3. Horizontal line
        hori_line_xstart, hori_line_xend = None, None

        if num_children == 1:
            hori_line_xstart = hori_line_xend = margin_vert_dash_below[0]
        elif num_children > 1:
            hori_line_xstart = margin_vert_dash_below[0]
            hori_line_xend = margin_vert_dash_below[-1]

        # STEP 4. Margin of current node
        margin_key_center = tmp = max(0, len(node.key) - 1) // 2

        if not is_leaf:
            margin_key_center = (hori_line_xstart + hori_line_xend) // 2

        margin_key = margin_key_center - tmp

        # STEP 5. Width of current node
        width_chbrsp = margin_prefix_children[-1] + width_children[-1] if not is_leaf else 0
        full_width = width_chbrsp if not is_leaf else len(node.key)

        # FINAL STEP
        node.width = full_width
        node.margin_key = margin_key
        node.margin_key_center = margin_key_center
        node.is_leaf = is_leaf

        if not is_leaf:
            node.width_chbrsp = width_chbrsp  # width of children + branch spacing
            node.hori_line_xstart = hori_line_xstart
            node.hori_line_xend = hori_line_xend
            node.margin_vert_dash_below = margin_vert_dash_below
            node.margin_prefix_children = margin_prefix_children

        self.adjust_too_long_key(node)

        del input_key
        del input_children
        del children
        del num_children
        del is_leaf
        del width_children
        del margin_prefix_children
        del margin_vert_dash_below
        del hori_line_xstart
        del hori_line_xend
        del margin_key
        del margin_key_center
        del width_chbrsp
        del full_width
        del tmp

        return node

    #
    #
    def get_margin_prefix_children(self, num_children: int, width_children: list):
        a = [0] * num_children

        for i in range(1, num_children):
            a[i] = a[i - 1] + width_children[i - 1] + self.space_branch_neighbor

        return a

    #
    #
    def get_margin_vert_dash_below(self, lst_children: list, margin_prefix_children: list):
        a = []

        for i in range(len(lst_children)):
            a.append(lst_children[i].margin_key_center + margin_prefix_children[i])

        return a

    #
    #
    def adjust_too_long_key(self, node: ParsingNode):
        len_key = len(node.key)

        if node.margin_key >= 0 and node.margin_key + len_key <= node.width:
            return

        shift_factor = 0 if node.margin_key >= 0 else abs(node.margin_key)

        node.width = max(node.width, node.margin_key + len_key)
        node.width += shift_factor

        self.shift_margin_except_children(node, shift_factor)

        del shift_factor

    #
    #
    def convert_margin_local_to_global(self, node: ParsingNode, shift_factor: int):
        self.shift_margin_except_children(node, shift_factor)

        for i in range(len(node.children)):
            child = node.children[i]
            child_shift_factor = node.margin_prefix_children[i]
            self.convert_margin_local_to_global(child, child_shift_factor)

    #
    #
    def shift_margin_except_children(self, node: ParsingNode, shift_factor: int):
        node.margin_key += shift_factor
        node.margin_key_center += shift_factor

        if not node.is_leaf:
            node.hori_line_xstart += shift_factor
            node.hori_line_xend += shift_factor
            node.margin_vert_dash_below = [value + shift_factor for value in node.margin_vert_dash_below]
            node.margin_prefix_children = [value + shift_factor for value in node.margin_prefix_children]

    #
    #
    def destroy_tree(self, node: ParsingNode):
        if node is None:
            return

        for child in node.children:
            self.destroy_tree(child)

        del node.key
        del node.children
        del node.width
        del node.margin_key
        del node.margin_key_center

        if not node.is_leaf:
            del node.width_chbrsp
            del node.hori_line_xstart
            del node.hori_line_xend
            del node.margin_vert_dash_below
            del node.margin_prefix_children

        del node.is_leaf

    #
    #
    def get_height(self, node) -> int:
        if node is None:
            return 0

        children = getattr(node, self.struct_node_children)
        max_height = 0

        for child in children:
            child_height = self.get_height(child)
            max_height = max(max_height, child_height)

        del children

        return 1 + max_height



class BinTreeIterIn:
    '''
    Binary Tree Interator.

    This class uses in-order traversal.
    '''
    def __init__(self, tree):
        self.__tree = tree
        self.__stack = []
        self.__node = tree._root

    #
    #
    def __next__(self):
        stack = self.__stack
        node = self.__node

        while node is not None:
            stack.append(node)
            node = node.left

        if not stack:
            raise StopIteration()

        res_node = stack.pop()
        self.__node = res_node.right

        del stack
        del node

        return res_node.key



from .abtreedisplay import AbstractTreeDisplay
from .valueutil import ValueUtil
from .matrixbuffer import MatrixBuffer
from .parsingnode import ParsingNode
from .bintreeparser import BinTreeParser


#
#
class BinTreeDisplay(AbstractTreeDisplay):
    '''
    A utility help visualize binary trees by using ASCII text.
    '''
    #
    #
    #################################################################
    #                        METHODS (PUBLIC)
    #################################################################
    #
    #
    def __init__(self):
        super().__init__()
        self._parser = BinTreeParser(self._vutil)
        self.config(compact_vert_line=True)

    #
    #
    def config(self, **kwargs):
        '''
        Configures settings.

        Args:
            struct_node: Tuple(name_key, name_left, name_right) indicating structure information of input node.

            space_branch_neighbor: Space between 2 branch neighbors.

            float_pre: Maximum precision of floating-point numbers.

            leaf_at_bottom: True if leaf will be drawn at bottom. Otherwise, False.

            compact_vert_line: True if vertical dash below will be omitted. Otherwise, False.

            margin_left: Left margin of output string result.

            hori_line_char: Display character for the horizontal line connecting branches.
        '''
        super()._config(kwargs)

        if kwargs:
            args_invalid = list(kwargs.keys())
            raise ValueError('Invalid argument(s):', args_invalid)

    #
    #
    #################################################################
    #                        METHODS (PROTECTED)
    #################################################################
    #
    #
    def _fill_buffer(self, node: ParsingNode, y_depth: int):
        if node is None:
            return

        if node.is_leaf:
            self._fill_buffer_node_leaf(node, y_depth)
            return

        self._buffer.fill_str(node.margin_key, y_depth, node.key)
        self._buffer.fill_str(node.margin_key_center, y_depth + 1, '|')
        self._buffer.fill_hori_line(self._hori_line_char, y_depth + 2, node.hori_line_xstart, node.hori_line_xend)

        y_depth += 3

        # vertical dash below
        if node.le is not None and not self._compact_vert_line:
            self._buffer.fill_str(node.margin_vert_dash_below_le, y_depth, '|')

        if node.ri is not None and not self._compact_vert_line:
            self._buffer.fill_str(node.margin_vert_dash_below_ri, y_depth, '|')

        if not self._compact_vert_line:
            y_depth += 1

        # process children
        if node.le is not None:
            self._fill_buffer(node.le, y_depth)

        if node.ri is not None:
            self._fill_buffer(node.ri, y_depth)

    #
    #
    def _get_height_buffer(self, height_inp_root: int) -> int:
        if self._compact_vert_line:
            return height_inp_root * 3 - 2

        return height_inp_root * 4 - 3
